from __future__ import print_function, division, absolute_import
from future import standard_library
from future.utils import iteritems

standard_library.install_aliases()
import inspect
import warnings
from array import array
from collections import defaultdict, deque
from importlib import import_module
from itertools import islice

from birdseye.utils import safe_qualname, correct_type, PY2, PY3

repr_registry = {}


def try_register_repr(module_name, class_name):
    try:
        cls = getattr(import_module(module_name), class_name)
    except (ImportError, AttributeError):
        return lambda x: x
    else:
        if inspect.isclass(cls):
            return register_repr(cls)


def register_repr(cls):
    def decorator(func):
        repr_registry[cls] = func
        return func

    return decorator


def maxparts(num):
    def decorator(func):
        func.maxparts = num
        return func

    return decorator


def basic_repr(x, *_):
    return '<%s instance at %#x>' % (type_name(x), id(x))


def type_name(x):
    return safe_qualname(correct_type(x))


"""
If a class has a __repr__ that returns a value longer than the threshold
below it will be noted as suppressed and not used again, and a warning
will be emitted.
"""
suppression_threshold = 300
supressed_classes = set()


class ReprSupressedWarning(Warning):
    pass


@register_repr(object)
@maxparts(60)
def repr_object(x, helper):
    try:
        s = repr(x)
        # Bugs in x.__repr__() can cause arbitrary
        # exceptions -- then make up something
    except Exception:
        return basic_repr(x)
    if len(s) > suppression_threshold:
        cls = correct_type(x)
        supressed_classes.add(cls)
        warnings.warn('%s.__repr__ is too long and has been supressed. '
                      'Register a cheap repr for the class to avoid this warning '
                      'and see an informative repr again.' % safe_qualname(cls))
    return helper.truncate(s)


class Repr(object):
    def __init__(self):
        self.maxlevel = 3

    def repr(self, x, level=None):
        if level is None:
            level = self.maxlevel
        for cls in inspect.getmro(correct_type(x)):
            if cls in supressed_classes:
                return basic_repr(x)[:-1] + ' (expensive repr suppressed)>'
            func = repr_registry.get(cls)
            if func:
                return func(x, ReprHelper(self, level, func))
        return repr(x)



aRepr = Repr()
cheap_repr = aRepr.repr


class ReprHelper(object):
    def __init__(self, repr_instance, level, func):
        self.repr_instance = repr_instance
        self.level = level
        self.func = func

    def repr_iterable(self, iterable, left, right, trail='', length=None):
        if length is None:
            length = len(iterable)
        if self.level <= 0 and length:
            s = '...'
        else:
            newlevel = self.level - 1
            repr1 = self.repr_instance.repr
            max_parts = self.maxparts
            pieces = [repr1(elem, newlevel) for elem in islice(iterable, max_parts)]
            if length > max_parts:
                pieces.append('...')
            s = ', '.join(pieces)
            if length == 1 and trail:
                right = trail + right
        return left + s + right

    @property
    def maxparts(self):
        return getattr(self.func, 'maxparts', 6)

    def truncate(self, s):
        max_parts = self.maxparts
        if len(s) > max_parts:
            i = max(0, (max_parts - 3) // 2)
            j = max(0, max_parts - 3 - i)
            s = s[:i] + '...' + s[len(s) - j:]
        return s


@register_repr(tuple)
def repr_tuple(x, helper):
    return helper.repr_iterable(x, '(', ')', trail=',')


@register_repr(list)
@try_register_repr('UserList', 'UserList')
@try_register_repr('collections', 'UserList')
def repr_list(x, helper):
    return helper.repr_iterable(x, '[', ']')


@register_repr(array)
@maxparts(5)
def repr_array(x, helper):
    if not x:
        return repr(x)
    return helper.repr_iterable(x, "array('%s', [" % x.typecode, '])')


@register_repr(set)
def repr_set(x, helper):
    if not x:
        return repr(x)
    if PY2:
        return helper.repr_iterable(x, 'set([', '])')
    else:
        return helper.repr_iterable(x, '{', '}')


@register_repr(frozenset)
def repr_frozenset(x, helper):
    if not x:
        return repr(x)
    if PY2:
        return helper.repr_iterable(x, 'frozenset([', '])')
    else:
        return helper.repr_iterable(x, 'frozenset({', '})')


@register_repr(deque)
def repr_deque(x, helper):
    return helper.repr_iterable(x, 'deque([', '])')


@register_repr(dict)
@try_register_repr('UserDict', 'UserDict')
@try_register_repr('collections', 'UserDict')
@maxparts(4)
def repr_dict(x, helper):
    n = len(x)
    if n == 0:
        return '{}'
    if helper.level <= 0:
        return '{...}'
    newlevel = helper.level - 1
    repr1 = helper.repr_instance.repr
    pieces = []
    for key in islice(x, helper.maxparts):
        keyrepr = repr1(key, newlevel)
        valrepr = repr1(x[key], newlevel)
        pieces.append('%s: %s' % (keyrepr, valrepr))
    if n > helper.maxparts:
        pieces.append('...')
    s = ', '.join(pieces)
    return '{%s}' % (s,)


@try_register_repr('__builtin__', 'unicode')
@register_repr(str)
@maxparts(30)
def repr_str(x, helper):
    return repr(helper.truncate(x))


@try_register_repr('builtins', 'int')
@try_register_repr('__builtin__', 'long')
@maxparts(40)
def repr_int(x, helper):
    return helper.truncate(repr(x))


@try_register_repr('numpy.core.multiarray', 'ndarray')
def repr_ndarray(x, helper):
    if len(x) == 0:
        return repr(x)
    return helper.repr_iterable(x, 'array([', '])')


@try_register_repr('django.db.models', 'QuerySet')
def repr_QuerySet(x, _):
    try:
        model_name = x.model._meta.object_name
    except AttributeError:
        model_name = type_name(x.model)
    return '<%s instance of %s at %#x>' % (type_name(x), model_name, id(x))


@try_register_repr('collections', 'ChainMap')
@try_register_repr('chainmap', 'ChainMap')
@maxparts(4)
def repr_ChainMap(x, helper):
    return helper.repr_iterable(x.maps, type_name(x) + '(', ')')


@try_register_repr('collections', 'OrderedDict')
@try_register_repr('ordereddict', 'OrderedDict')
@try_register_repr('backport_collections', 'OrderedDict')
@maxparts(4)
def repr_OrderedDict(x, helper):
    helper.level += 1
    return helper.repr_iterable(iteritems(x), type_name(x) + '(', ')', length=len(x))


@try_register_repr('collections', 'Counter')
@try_register_repr('counter', 'Counter')
@try_register_repr('backport_collections', 'Counter')
@maxparts(5)
def repr_Counter(x, helper):
    typename = type_name(x)
    if not x:
        return '%s()' % typename
    return '{0}({1})'.format(typename, repr_dict(x, helper))


@register_repr(defaultdict)
@maxparts(4)
def repr_defaultdict(x, helper):
    return '{0}({1}, {2})'.format(type_name(x),
                                  x.default_factory,
                                  repr_dict(x, helper))



from __future__ import print_function, division, absolute_import

from future import standard_library

standard_library.install_aliases()
import json
import os

from humanize import naturaltime
from markupsafe import Markup
from sqlalchemy import Sequence, UniqueConstraint, create_engine, Column, Integer, Text, ForeignKey, DateTime
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import backref, relationship, sessionmaker
from sqlalchemy.pool import StaticPool


DB_URI = os.environ.get('BIRDSEYE_DB',
                        'sqlite:///' + os.path.join(os.path.expanduser('~'),
                                                    '.birdseye.db'))

engine = create_engine(DB_URI,
                       connect_args={'check_same_thread': False},
                       poolclass=StaticPool)

Session = sessionmaker(bind=engine)
session = Session()


class Base(object):
    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()


Base = declarative_base(cls=Base)  # type: ignore


class Call(Base):
    id = Column(Text(length=32), primary_key=True)
    function_id = Column(Integer, ForeignKey('function.id'))
    function = relationship('Function', backref=backref('calls', lazy='dynamic'))
    arguments = Column(Text)
    return_value = Column(Text)
    exception = Column(Text)
    traceback = Column(Text)
    data = Column(Text)
    start_time = Column(DateTime)

    @property
    def pretty_start_time(self):
        return self._pretty_time(self.start_time)

    @staticmethod
    def _pretty_time(dt):
        if not dt:
            return ''
        return Markup('%s (%s)' % (
            dt.strftime('%Y-%m-%d&nbsp;%H:%M:%S'),
            naturaltime(dt)))

    @property
    def state_icon(self):
        return Markup('<span class="glyphicon glyphicon-%s" '
                      'style="color: %s"></span>' % (
                          ('ok', 'green') if self.success else
                          ('remove', 'red')))

    @property
    def success(self):
        if self.exception:
            assert self.traceback
            assert self.return_value == 'None'
            return False
        else:
            assert not self.traceback
            return True

    @property
    def result(self):
        if self.success:
            return self.return_value
        else:
            return self.exception

    @property
    def arguments_list(self):
        return json.loads(self.arguments)


class Function(Base):
    id = Column(Integer, Sequence('function_id_seq'), primary_key=True)
    file = Column(Text)
    name = Column(Text)
    html_body = Column(Text)
    lineno = Column(Integer)
    data = Column(Text)

    __table_args__ = (UniqueConstraint('file', 'name', 'html_body', 'lineno', 'data',
                                       name='everything_unique'),)


Base.metadata.create_all(engine)



from __future__ import print_function, division, absolute_import
from future import standard_library

standard_library.install_aliases()
from birdseye.server import main

if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

from future import standard_library

standard_library.install_aliases()
from future.utils import iteritems
from typing import List, Dict, Any, Optional, NamedTuple, Tuple, Iterator, Iterable, Union, cast
from types import FrameType, TracebackType, CodeType, FunctionType
import typing

import ast
import html
import inspect
import json
import os
import traceback
from collections import defaultdict, Sequence, Set, Mapping, deque
from datetime import datetime
from functools import partial
from itertools import chain, islice
from threading import Lock
from uuid import uuid4

from asttokens import ASTTokens
from littleutils import group_by_key_func

from birdseye.cheap_repr import cheap_repr
from birdseye.db import Function, Call, session
from birdseye.tracer import TreeTracerBase, TracedFile, EnterCallInfo, ExitCallInfo, FrameInfo, ChangeValue
from birdseye import tracer
from birdseye.utils import safe_qualname, correct_type, exception_string, dummy_namespace, PY3, PY2, one_or_none, \
    of_type, Deque, Text, flatten_list

CodeInfo = NamedTuple('CodeInfo', [('db_func', Function),
                                   ('traced_file', TracedFile)])


class BirdsEye(TreeTracerBase):
    def __init__(self):
        super(BirdsEye, self).__init__()
        self._code_infos = {}  # type: Dict[CodeType, CodeInfo]

    def parse_extra(self, root, source, filename):
        for node in ast.walk(root):  # type: ast.AST
            node._loops = tracer.loops(node)
            if isinstance(node, ast.expr):
                node._is_interesting_expression = is_interesting_expression(node)

    def compile(self, source, filename):
        traced_file = super(BirdsEye, self).compile(source, filename)
        traced_file.tokens = ASTTokens(source, tree=traced_file.root)
        return traced_file

    def before_stmt(self, node, frame):
        # type: (ast.stmt, FrameType) -> None
        if frame.f_code not in self._code_infos:
            return
        if isinstance(node.parent, (ast.For, ast.While)) and node is node.parent.body[0]:
            self._add_iteration(node._loops, frame)
        self._set_node_value(node, frame, True)

    def _add_iteration(self, loops, frame):
        # type: (typing.Sequence[ast.AST], FrameType) -> None
        iteration = self.stack[frame].iteration  # type: Iteration
        for i, loop_node in enumerate(loops):
            loop = iteration.loops[loop_node._tree_index]
            if i == len(loops) - 1:
                loop.append(Iteration())
            else:
                iteration = loop.last()

    def before_expr(self, node, frame):
        # type: (ast.expr, FrameType) -> None
        self.stack[frame].inner_call = None

    def after_expr(self, node, frame, value):
        # type: (ast.expr, FrameType, Any) -> Optional[ChangeValue]
        if node._is_interesting_expression:
            original_frame = frame
            while frame.f_code.co_name in ('<listcomp>',
                                           '<dictcomp>',
                                           '<setcomp>'):
                frame = frame.f_back

            if frame.f_code not in self._code_infos:
                return None

            if is_obvious_builtin(node, self.stack[original_frame]):
                return None

            frame_info = self.stack[frame]
            expanded_value = expand(value, level=max(1, 3 - len(node._loops)))
            if frame_info.inner_call:
                expanded_value.insert(2, {'inner_call': frame_info.inner_call})
                frame_info.inner_call = None
            self._set_node_value(node, frame, expanded_value)

        is_special_comprehension_iter = (isinstance(node.parent, ast.comprehension) and
                                         node is node.parent.iter and
                                         not isinstance(node.parent.parent, ast.GeneratorExp))
        if not is_special_comprehension_iter:
            return None

        self._set_node_value(node.parent, frame, True)

        def comprehension_iter_proxy():
            loops = node._loops + (node.parent,)  # type: Tuple[ast.AST, ...]
            for item in value:
                self._add_iteration(loops, frame)
                yield item

        return ChangeValue(comprehension_iter_proxy())

    def _set_node_value(self, node, frame, value):
        # type: (ast.AST, FrameType, Any) -> None
        iteration = self.stack[frame].iteration  # type: Iteration
        for i, loop_node in enumerate(node._loops):  # type: int, ast.AST
            loop = iteration.loops[loop_node._tree_index]
            iteration = loop.last()
        iteration.vals[node._tree_index] = value

    def after_stmt(self, node, frame, exc_value, exc_traceback):
        # type: (ast.stmt, FrameType, Exception, TracebackType) -> None
        if frame.f_code not in self._code_infos:
            return
        frame_info = self.stack[frame]
        expression_stack = frame_info.expression_stack
        if expression_stack:
            while isinstance(expression_stack[-1], self.SPECIAL_COMPREHENSION_TYPES):
                inner_frame = frame_info.comprehension_frames[expression_stack[-1]]
                expression_stack = self.stack[inner_frame].expression_stack
            self._set_node_value(
                expression_stack[-1], frame,
                [exception_string(exc_value), -1])

    def enter_call(self, enter_info):
        # type: (EnterCallInfo) -> None
        frame = enter_info.current_frame  # type: FrameType
        if frame.f_code not in self._code_infos:
            return
        frame_info = self.stack[frame]
        frame_info.start_time = datetime.now()
        frame_info.iteration = Iteration()
        arg_info = inspect.getargvalues(frame)
        # TODO keep argument names in code info
        arg_names = chain(flatten_list(arg_info[0]), arg_info[1:3])  # type: Iterator[str]
        f_locals = arg_info[3].copy()  # type: Dict[str, Any]
        arguments = [(name, f_locals.pop(name)) for name in arg_names if name] + [
            it for it in f_locals.items()
            if it[0][0] != '.'  # Appears when using nested tuple arguments
        ]
        frame_info.arguments = json.dumps([[k, cheap_repr(v)] for k, v in arguments])
        self.stack.get(frame.f_back, dummy_namespace).inner_call = frame_info.call_id = self._call_id()

    def _call_id(self):
        # type: () -> Text
        return uuid4().hex

    def exit_call(self, exit_info):
        # type: (ExitCallInfo) -> None
        frame = exit_info.current_frame  # type: FrameType
        if frame.f_code not in self._code_infos:
            return
        frame_info = self.stack[frame]
        top_iteration = frame_info.iteration  # type: Iteration

        loop_iterations = top_iteration.extract_iterations()['loops']

        node_values = _deep_dict()

        def extract_values(iteration, path):
            # type: (Iteration, Tuple[int, ...]) -> None
            for k, v in iteration.vals.items():
                full_path = (k,) + path
                d = node_values
                for path_k in full_path[:-1]:
                    d = d[path_k]
                d[full_path[-1]] = v

            for loop in iteration.loops.values():
                for i, iteration in enumerate(loop):
                    extract_values(iteration, path + (i,))

        extract_values(top_iteration, ())

        db_func = self._code_infos[frame.f_code].db_func  # type: Function
        exc = exit_info.exc_value  # type: Optional[Exception]
        if exc:
            traceback_str = ''.join(traceback.format_exception(type(exc), exc, exit_info.exc_tb))
            exception = exception_string(exc)
        else:
            traceback_str = exception = None

        call = Call(id=frame_info.call_id,
                    function=db_func,
                    arguments=frame_info.arguments,
                    return_value=cheap_repr(exit_info.return_value),
                    exception=exception,
                    traceback=traceback_str,
                    data=json.dumps(
                        dict(
                            node_values=node_values,
                            loop_iterations=loop_iterations,
                            type_names=type_registry.names(),
                            num_special_types=type_registry.num_special_types,
                        ),
                        separators=(',', ':')
                    ),
                    start_time=frame_info.start_time)
        session.add(call)
        session.commit()

    def __call__(self, func_or_class):
        # type: (Union[FunctionType, type]) -> (Union[FunctionType, type])
        if inspect.isclass(func_or_class):
            cls = cast(type, func_or_class)
            for name, meth in iteritems(cls.__dict__):  # type: str, FunctionType
                if inspect.ismethod(meth) or inspect.isfunction(meth):
                    setattr(cls, name, self.__call__(meth))
            return cls

        func = cast(FunctionType, func_or_class)
        new_func = super(BirdsEye, self).__call__(func)
        code_info = self._code_infos.get(new_func.__code__)
        if code_info:
            return new_func
        lines, start_lineno = inspect.getsourcelines(func)  # type: List[Text], int
        end_lineno = start_lineno + len(lines)
        name = safe_qualname(func)
        filename = os.path.abspath(inspect.getsourcefile(func))

        traced_file = new_func.traced_file  # type: TracedFile
        traced_file.root._depth = 0
        for node in ast.walk(traced_file.root):  # type: ast.AST
            for child in ast.iter_child_nodes(node):
                child._depth = node._depth + 1

        positions = []  # type: List[Tuple[int, int, int, str]]
        node_loops = {}  # type: Dict[int, List[int]]
        for node in traced_file.nodes:
            if isinstance(node, ast.expr):
                node_type = 'expr'
                if not node._is_interesting_expression:
                    continue
            elif (isinstance(node, (ast.While, ast.For, ast.comprehension))
                  and not isinstance(node.parent, ast.GeneratorExp)):
                node_type = 'loop'
            elif isinstance(node, ast.stmt):
                node_type = 'stmt'
            else:
                continue
            assert isinstance(node, ast.AST)

            # In particular FormattedValue is missing this
            if not hasattr(node, 'first_token'):
                continue

            if not start_lineno <= node.first_token.start[0] <= end_lineno:
                continue

            start, end = traced_file.tokens.get_text_range(node)  # type: int, int
            if start == end == 0:
                continue
            positions.append((start, 1, node._depth,
                              '<span data-index="%s" data-type="%s">' % (node._tree_index, node_type)))
            positions.append((end, 0, node._depth, '</span>'))
            if node._loops:
                node_loops[node._tree_index] = [n._tree_index for n in node._loops]

        comprehensions = group_by_key_func(of_type(ast.comprehension, traced_file.nodes),
                                           lambda c: c.first_token.line
                                           )  # type: Dict[Any, Iterable[ast.comprehension]]

        def get_start(n):
            # type: (ast.AST) -> int
            return traced_file.tokens.get_text_range(n)[0]

        for comp_list in comprehensions.values():
            prev_start = None  # type: Optional[int]
            for comp in sorted(comp_list, key=lambda c: c.first_token.startpos):
                if comp is comp.parent.generators[0]:
                    start = get_start(comp.parent)
                    if prev_start is not None and start < prev_start:
                        start = get_start(comp)
                else:
                    start = get_start(comp)
                if prev_start is not None:
                    positions.append((start, 1, 0, '\n '))
                    end_lineno += 1
                prev_start = start

        positions.append((len(traced_file.source), 0, 0, ''))
        positions.sort()

        html_lines = []
        start = 0
        for pos, _, _, part in positions:
            html_lines.append(html.escape(traced_file.source[start:pos]))
            html_lines.append(part)
            start = pos
        html_body = ''.join(html_lines)
        html_body = '\n'.join(html_body.split('\n')[start_lineno - 1:end_lineno - 1])

        db_args = dict(file=filename,
                       name=name,
                       html_body=html_body,
                       lineno=start_lineno,
                       data=json.dumps(
                           dict(
                               node_loops=node_loops,
                           ),
                           sort_keys=True,
                       ))

        db_func = one_or_none(session.query(Function).filter_by(**db_args))  # type: Optional[Function]
        if not db_func:
            db_func = Function(**db_args)
            session.add(db_func)
            session.commit()
        self._code_infos[new_func.__code__] = CodeInfo(db_func, traced_file)

        return new_func


eye = BirdsEye()


def _deep_dict():
    return defaultdict(_deep_dict)


class Iteration(object):
    def __init__(self):
        self.vals = {}  # type: Dict[int, Any]
        self.loops = defaultdict(IterationList)  # type: Dict[int, IterationList]
        self.index = None  # type: int

    def extract_iterations(self):
        # type: () -> Dict[str, Union[int, Dict]]
        return {
            'index': self.index,
            'loops': {
                k: [it.extract_iterations() for it in v]
                for k, v in self.loops.items()
            }
        }


class IterationList(Iterable[Iteration]):
    side_len = 3

    def __init__(self):
        self.start = []  # type: List[Iteration]
        self.end = deque(maxlen=self.side_len)  # type: Deque[Iteration]
        self.length = 0  # type: int

    def append(self, iteration):
        # type: (Iteration) -> None
        if self.length < self.side_len:
            self.start.append(iteration)
        else:
            self.end.append(iteration)
        iteration.index = self.length
        self.length += 1

    def __iter__(self):
        # type: () -> Iterator[Iteration]
        return chain(self.start, self.end)

    def last(self):
        # type: () -> Iteration
        if self.end:
            return self.end[-1]
        else:
            return self.start[-1]


class TypeRegistry(object):
    def __init__(self):
        self.lock = Lock()
        self.data = defaultdict(lambda: len(self.data))  # type: Dict[type, int]
        basic_types = [type(None), bool, int, float, complex]
        if PY2:
            basic_types += [long]
        special_types = basic_types + [list, dict, tuple, set, frozenset, str]
        if PY2:
            special_types += [unicode if PY2 else bytes]

        self.num_basic_types = len(basic_types)
        self.num_special_types = len(special_types)
        for t in special_types:
            _ = self.data[t]

    def __getitem__(self, item):
        t = correct_type(item)
        with self.lock:
            return self.data[t]

    def names(self):
        # type: () -> List[str]
        rev = dict((v, k) for k, v in self.data.items())
        return [safe_qualname(rev[i]) for i in range(len(rev))]


type_registry = TypeRegistry()


def expand(val, level=3):
    type_index = type_registry[val]
    result = [cheap_repr(val), type_index]
    if type_index < type_registry.num_basic_types or level == 0:
        return result
    exp = partial(expand, level=level - 1)

    # noinspection PyBroadException
    try:
        length = len(val)
    except:
        pass
    else:
        result += ['len() = %s' % length]

    if isinstance(val, (str, bytes, range) if PY3 else (str, unicode, xrange)):
        return result
    if isinstance(val, Sequence):
        if len(val) <= 8:
            indices = range(len(val))
        else:
            indices = chain(range(3), range(len(val) - 3, len(val)))
        for i in indices:
            result += [(str(i), exp(val[i]))]
    elif isinstance(val, Mapping):
        for k, v in islice(iteritems(val), 10):
            result += [(cheap_repr(k), exp(v))]
    elif isinstance(val, Set):
        if len(val) <= 8:
            vals = val
        else:
            vals = islice(val, 6)
        for i, v in enumerate(vals):
            result += [('<%s>' % i, exp(v))]

    d = getattr(val, '__dict__', None)
    if d:
        for k, v in islice(iteritems(d), 50):
            if isinstance(v, TracedFile):
                continue
            result += [(str(k), exp(v))]
    else:
        slots = getattr(val, '__slots__', None)
        if slots:
            for s in slots:
                try:
                    attr = getattr(val, s)
                except AttributeError:
                    pass
                else:
                    result += [(str(s), exp(attr))]
    return result


def is_interesting_expression(node):
    # type: (ast.AST) -> bool
    return (isinstance(node, ast.expr) and
            not (isinstance(node, (ast.Num, ast.Str, getattr(ast, 'NameConstant', ()))) or
                 isinstance(getattr(node, 'ctx', None),
                            (ast.Store, ast.Del)) or
                 (isinstance(node, ast.UnaryOp) and
                  isinstance(node.op, (ast.UAdd, ast.USub)) and
                  isinstance(node.operand, ast.Num)) or
                 (isinstance(node, (ast.List, ast.Tuple, ast.Dict)) and
                  not any(is_interesting_expression(n) for n in ast.iter_child_nodes(node)))))


def is_obvious_builtin(node, frame_info):
    # type: (ast.expr, FrameInfo) -> bool
    value = frame_info.expression_values[node]
    # noinspection PyUnresolvedReferences
    builtins = cast(dict, __builtins__)
    return ((isinstance(node, ast.Name) and
             node.id in builtins and
             builtins[node.id] is value) or
            isinstance(node, getattr(ast, 'NameConstant', ())))



from __future__ import print_function, division, absolute_import

from future import standard_library

standard_library.install_aliases()
from future.utils import raise_from
import atexit
import ntpath
import os
import traceback
import types
from queue import Queue
from threading import Thread
from sys import version_info
from typing import TypeVar, Union, List, Any, Iterator, Tuple, Iterable, Dict
from types import FunctionType

try:
    from typing import Type
except ImportError:
    Type = type

try:
    from typing import Deque
except ImportError:
    from collections import deque as Deque

from littleutils import strip_required_prefix
from qualname import qualname

PY2 = version_info.major == 2
PY3 = not PY2
T = TypeVar('T')
RT = TypeVar('RT')

if PY2:
    Text = unicode
else:
    Text = str


def path_leaf(path):
    # type: (str) -> str
    # http://stackoverflow.com/a/8384788/2482744
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def all_file_paths():
    # type: () -> List[str]
    from birdseye.db import Function, Session
    return [f[0] for f in Session().query(Function.file).distinct()]


def common_ancestor(paths):
    # type: (List[str]) -> str
    """
    Returns a path to a directory that contains all the given absolute paths
    """
    prefix = os.path.commonprefix(paths)

    # Ensure that the prefix doesn't end in part of the name of a file/directory
    prefix = ntpath.split(prefix)[0]

    # Ensure that it ends with a slash
    first_char_after = paths[0][len(prefix)]
    if first_char_after in r'\/':
        prefix += first_char_after

    return prefix


def short_path(path, all_paths=None):
    # type: (str, List[str]) -> str
    prefix = common_ancestor(all_paths or all_file_paths())
    if prefix in r'\/':
        prefix = ''
    return strip_required_prefix(path, prefix) or path_leaf(path)


def safe_qualname(obj):
    # type: (Union[Type, FunctionType]) -> str
    result = _safe_qualname_cache.get(obj)
    if not result:
        try:
            result = qualname(obj)
        except AttributeError:
            result = obj.__name__
        if '<locals>' not in result:
            _safe_qualname_cache[obj] = result
    return result


_safe_qualname_cache = {}  # type: Dict[Union[Type, FunctionType], str]


def correct_type(obj):
    # type: (Any) -> type
    # TODO handle case where __class__ has been assigned
    t = type(obj)
    if t is getattr(types, 'InstanceType', None):
        t = obj.__class__
    return t


def iter_get(it, n):
    n_original = n
    if n < 0:
        n = -n - 1
        it = reversed(it)
    else:
        it = iter(it)
    try:
        while n > 0:
            next(it)
            n -= 1
        return next(it)
    except StopIteration as e:
        raise_from(IndexError(n_original), e)


def exception_string(exc):
    # type: (BaseException) -> Text
    assert isinstance(exc, BaseException)
    return ''.join(traceback.format_exception_only(type(exc), exc))


class Consumer(object):
    def __init__(self):
        self.queue = Queue()
        self.error = ValueError('There should be an error raised by a consumed task here')
        self.exiting = False
        self._run_thread()
        atexit.register(self._exit)

    def _run_thread(self):
        def run():
            while True:
                func = self.queue.get()
                try:
                    func()
                except BaseException as e:
                    self.queue = None
                    self.error = e
                    if not self.exiting:
                        raise
                finally:
                    self.queue.task_done()

        self.thread = Thread(target=run, name='Consumer thread')
        self.thread.daemon = True
        self.thread.start()

    def _exit(self):
        self.exiting = True
        if self.queue:
            self._run_thread()  # just for good measure
            self.queue.join()

    def __call__(self, func):
        if self.queue:
            self.queue.put(func)
        else:
            raise self.error  # error raised by task in consumer thread

    def wait(self, func):
        self(func)
        self.queue.join()


class SimpleNamespace(object):
    pass


dummy_namespace = SimpleNamespace()


def of_type(type_or_tuple, iterable):
    # type: (Union[type, Tuple[Union[type, tuple], ...]], Iterable[Any]) -> Iterator[Any]
    return (x for x in iterable if isinstance(x, type_or_tuple))


def safe_next(it):
    # type: (Iterator[T]) -> T
    """
    next() can raise a StopIteration which can cause strange bugs inside generators.
    """
    try:
        return next(it)
    except StopIteration as e:
        raise_from(RuntimeError, e)
        raise  # isn't reached


def one_or_none(expression):
    """Performs a one_or_none on a sqlalchemy expression."""
    if hasattr(expression, 'one_or_none'):
        return expression.one_or_none()
    result = expression.all()
    if len(result) == 0:
        return None
    elif len(result) == 1:
        return result[0]
    else:
        raise Exception("There is more than one item returned for the supplied filter")


def flatten_list(lst):
    result = []
    for x in lst:
        if isinstance(x, list):
            result.extend(flatten_list(x))
        else:
            result.append(x)
    return result



from __future__ import print_function, division, absolute_import

from future import standard_library

standard_library.install_aliases()

import ast
import inspect
from collections import namedtuple
from copy import deepcopy
from functools import partial, update_wrapper
from itertools import takewhile
from typing import List, Dict, Any, Optional, NamedTuple, Tuple, Iterator, Callable, cast
from types import FrameType, TracebackType, CodeType, FunctionType

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache
from littleutils import file_to_string

from birdseye.utils import of_type, safe_next, PY3, T, Type


class TracedFile(object):
    def __init__(self, tracer, source, filename):
        # type: (TreeTracerBase, str, str) -> None
        root = ast.parse(source, filename)  # type: ast.Module

        self.nodes = []  # type: List[ast.AST]

        def set_basic_node_attributes():
            self.nodes = []  # type: List[ast.AST]
            for node in ast.walk(root):  # type: ast.AST
                for child in ast.iter_child_nodes(node):
                    child.parent = node
                node.traced_file = self
                node._tree_index = len(self.nodes)
                self.nodes.append(node)

        set_basic_node_attributes()

        new_root = tracer.parse_extra(root, source, filename)
        if new_root is not None:
            root = new_root

        set_basic_node_attributes()

        new_root = deepcopy(root)
        new_root = _NodeVisitor().visit(new_root)

        self.code = compile(new_root, filename, "exec")  # type: CodeType
        self.root = root
        self.tracer = tracer
        self.source = source
        self.filename = filename


class FrameInfo(object):
    def __init__(self):
        self.statement = None  # type: Optional[ast.stmt]
        self.expression_stack = []  # type: List[ast.expr]
        self.expression_values = {}  # type: Dict[ast.expr, Any]
        self.return_node = None  # type: Optional[ast.Return]
        self.comprehension_frames = {}  # type: Dict[ast.expr, FrameType]


EnterCallInfo = NamedTuple('EnterCallInfo', [('call_node', Optional[ast.expr]),
                                             ('enter_node', ast.AST),
                                             ('caller_frame', FrameType),
                                             ('current_frame', FrameType)])
ExitCallInfo = NamedTuple('ExitCallInfo', [('call_node', Optional[ast.expr]),
                                           ('return_node', Optional[ast.Return]),
                                           ('caller_frame', FrameType),
                                           ('current_frame', FrameType),
                                           ('return_value', Any),
                                           ('exc_value', Optional[Exception]),
                                           ('exc_tb', Optional[TracebackType])])

ChangeValue = namedtuple('ChangeValue', 'value')


class TreeTracerBase(object):
    SPECIAL_COMPREHENSION_TYPES = (ast.DictComp, ast.SetComp)  # type: Tuple[Type[ast.expr], ...]
    if PY3:
        SPECIAL_COMPREHENSION_TYPES += (ast.ListComp,)

    def __init__(self):
        self.stack = {}  # type: Dict[FrameType, FrameInfo]

    @lru_cache()
    def compile(self, source, filename):
        # type: (str, str) -> TracedFile
        return TracedFile(self, source, filename)

    def exec_string(self, source, filename, globs=None, locs=None):
        # type: (str, str, dict, dict) -> None
        traced_file = self.compile(source, filename)
        globs = globs or {}
        locs = locs or {}
        globs = dict(globs, **self._trace_methods_dict(traced_file))
        exec (traced_file.code, globs, locs)

    def _trace_methods_dict(self, traced_file):
        # type: (TracedFile) -> Dict[str, Callable]
        return {f.__name__: partial(f, traced_file)
                for f in [
                    self._treetrace_hidden_with_stmt,
                    self._treetrace_hidden_before_expr,
                    self._treetrace_hidden_after_expr,
                ]}

    def __call__(self, func):
        # type: (FunctionType) -> FunctionType
        try:
            if inspect.iscoroutinefunction(func) or inspect.isasyncgenfunction(func):
                raise ValueError('You cannot trace async functions')
        except AttributeError:
            pass
        filename = inspect.getsourcefile(func)  # type: str
        source = file_to_string(filename)
        traced_file = self.compile(source, filename)
        func.__globals__.update(self._trace_methods_dict(traced_file))

        code_options = []  # type: List[CodeType]

        def find_code(root_code):
            # type: (CodeType) -> None
            for const in root_code.co_consts:  # type: CodeType
                if not inspect.iscode(const):
                    continue
                matches = (const.co_firstlineno == func.__code__.co_firstlineno and
                           const.co_name == func.__code__.co_name)
                if matches:
                    code_options.append(const)
                find_code(const)

        find_code(traced_file.code)
        if len(code_options) > 1:
            assert func.__code__.co_name == (lambda: 0).__code__.co_name
            raise ValueError("Failed to trace lambda. Convert the function to a def.")
        new_func_code = code_options[0]  # type: CodeType

        # http://stackoverflow.com/a/13503277/2482744
        # noinspection PyArgumentList
        new_func = FunctionType(new_func_code, func.__globals__, func.__name__, func.__defaults__, func.__closure__)
        update_wrapper(new_func, func)  # type: FunctionType
        if PY3:
            new_func.__kwdefaults__ = getattr(func, '__kwdefaults__', None)
        new_func.traced_file = traced_file
        return new_func

    def _treetrace_hidden_with_stmt(self, traced_file, _tree_index):
        # type: (TracedFile, int) -> _StmtContext
        node = traced_file.nodes[_tree_index]
        node = cast(ast.stmt, node)
        frame = inspect.currentframe().f_back  # type: FrameType
        return _StmtContext(self, node, frame)

    def _treetrace_hidden_before_expr(self, traced_file, _tree_index):
        # type: (TracedFile, int) -> ast.expr
        node = traced_file.nodes[_tree_index]
        node = cast(ast.expr, node)
        frame = inspect.currentframe().f_back  # type: FrameType

        frame_info = self.stack.get(frame)
        if frame_info is None:
            frame_info = FrameInfo()
            self.stack[frame] = frame_info
            owner_frame = frame
            while owner_frame.f_code.co_name in ('<listcomp>', '<dictcomp>', '<setcomp>'):
                owner_frame = owner_frame.f_back
            if owner_frame != frame:
                comprehension = safe_next(of_type(self.SPECIAL_COMPREHENSION_TYPES,
                                                  ancestors(node)))  # type: ast.expr
                self.stack[owner_frame].comprehension_frames[comprehension] = frame

        frame_info.expression_stack.append(node)

        self.before_expr(node, frame)
        return node

    def _treetrace_hidden_after_expr(self, _, node, value):
        # type: (TracedFile, ast.expr, Any) -> Any
        frame = inspect.currentframe().f_back  # type: FrameType
        self.stack[frame].expression_stack.pop()
        self.stack[frame].expression_values[node] = value
        result = self.after_expr(node, frame, value)
        if result is not None:
            assert isinstance(result, ChangeValue), "after_expr must return None or an instance of ChangeValue"
            value = result.value
        return value

    def _enter_call(self, enter_node, current_frame):
        # type: (ast.AST, FrameType) -> None
        caller_frame, call_node = self._get_caller_stuff(current_frame)
        self.stack[current_frame] = FrameInfo()
        self.enter_call(EnterCallInfo(call_node, enter_node, caller_frame, current_frame))

    def _get_caller_stuff(self, frame):
        # type: (FrameType) -> Tuple[FrameType, Optional[ast.expr]]
        caller_frame = frame.f_back
        call_node = None
        if caller_frame in self.stack:
            expression_stack = self.stack[caller_frame].expression_stack
            if expression_stack:
                call_node = expression_stack[-1]
        return caller_frame, call_node

    def before_expr(self, node, frame):
        # type: (ast.expr, FrameType) -> None
        pass

    def after_expr(self, node, frame, value):
        # type: (ast.expr, FrameType, Any) -> Optional[ChangeValue]
        pass

    def before_stmt(self, node, frame):
        # type: (ast.stmt, FrameType) -> None
        pass

    def after_stmt(self, node, frame, exc_value, exc_traceback):
        # type: (ast.stmt, FrameType, Exception, TracebackType) -> Optional[bool]
        pass

    def enter_call(self, enter_info):
        # type: (EnterCallInfo) -> None
        pass

    def exit_call(self, exit_info):
        # type: (ExitCallInfo) -> None
        pass

    def parse_extra(self, root, source, filename):
        # type: (ast.Module, str, str) -> Optional[ast.Module]
        pass


class _NodeVisitor(ast.NodeTransformer):
    def generic_visit(self, node):
        # type: (ast.AST) -> ast.AST
        if isinstance(node, ast.expr) and not (hasattr(node, "ctx") and not isinstance(node.ctx, ast.Load)):
            return self.visit_expr(node)
        if isinstance(node, ast.stmt):
            if not (isinstance(node, ast.ImportFrom) and node.module == "__future__"):
                return self.visit_stmt(node)
        return super(_NodeVisitor, self).generic_visit(node)

    def visit_expr(self, node):
        # type: (ast.expr) -> ast.Call
        """
        each expression e gets wrapped like this:
            _after(_before(_tree_index), e)
        where
            _after is function that gives the resulting value
            _before is function that signals the beginning of evaluation of e
        """

        if isinstance(node, getattr(ast, 'Starred', ())):
            return super(_NodeVisitor, self).generic_visit(node)

        before_marker = _create_simple_marker_call(node, TreeTracerBase._treetrace_hidden_before_expr)
        ast.copy_location(before_marker, node)

        after_marker = ast.Call(
            func=ast.Name(id=TreeTracerBase._treetrace_hidden_after_expr.__name__,
                          ctx=ast.Load()),
            args=[
                before_marker,
                super(_NodeVisitor, self).generic_visit(node),
            ],
            keywords=[],
        )
        ast.copy_location(after_marker, node)
        ast.fix_missing_locations(after_marker)

        return after_marker

    def visit_stmt(self, node):
        # type: (ast.stmt) -> ast.With
        context_expr = _create_simple_marker_call(
            super(_NodeVisitor, self).generic_visit(node),
            TreeTracerBase._treetrace_hidden_with_stmt)

        if PY3:
            wrapped = ast.With(
                items=[ast.withitem(context_expr=context_expr)],
                body=[node],
            )
        else:
            wrapped = ast.With(
                context_expr=context_expr,
                body=[node],
            )
        ast.copy_location(wrapped, node)
        ast.fix_missing_locations(wrapped)
        return wrapped


class _StmtContext(object):
    __slots__ = ('tracer', 'node', 'frame')

    def __init__(self, tracer, node, frame):
        # type: (TreeTracerBase, ast.stmt, FrameType) -> None
        self.tracer = tracer
        self.node = node
        self.frame = frame

    def __enter__(self):
        tracer = self.tracer
        node = self.node
        frame = self.frame
        if isinstance(node.parent, (ast.FunctionDef, ast.Module)) and node is node.parent.body[0]:
            tracer._enter_call(node, frame)
        frame_info = tracer.stack[frame]
        frame_info.expression_stack = []
        frame_info.statement = node
        tracer.before_stmt(node, frame)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # type: (Type[Exception], Exception, TracebackType) -> bool
        node = self.node
        tracer = self.tracer
        frame = self.frame
        result = tracer.after_stmt(node, frame, exc_val, exc_tb)
        frame_info = tracer.stack[frame]
        if isinstance(node, ast.Return):
            frame_info.return_node = node
        parent = node.parent  # type: ast.AST
        return_node = frame_info.return_node
        exiting = (isinstance(parent, (ast.FunctionDef, ast.Module)) and
                   (node is parent.body[-1] or
                    exc_val or
                    return_node))
        if exiting:
            caller_frame, call_node = tracer._get_caller_stuff(frame)
            return_value = None
            if return_node and return_node.value and not exc_val:
                return_value = frame_info.expression_values[return_node.value]
            tracer.exit_call(ExitCallInfo(call_node,
                                          return_node,
                                          caller_frame,
                                          frame,
                                          return_value,
                                          exc_val,
                                          exc_tb
                                          ))

            del tracer.stack[frame]
            for comprehension_frame in frame_info.comprehension_frames.values():
                del tracer.stack[comprehension_frame]
        return result


def _create_simple_marker_call(node, func):
    # type: (ast.AST, Callable) -> ast.Call
    return ast.Call(
        func=ast.Name(id=func.__name__,
                      ctx=ast.Load()),
        args=[ast.Num(node._tree_index)],
        keywords=[],
    )


def ancestors(node):
    # type: (ast.AST) -> Iterator[ast.AST]
    while True:
        try:
            node = node.parent
        except AttributeError:
            break
        yield node


def loops(node):
    # type: (ast.AST) -> Tuple[ast.AST, ...]
    result = []
    while True:
        try:
            parent = node.parent
        except AttributeError:
            break
        if isinstance(parent, (ast.FunctionDef, ast.ClassDef)):
            break

        is_containing_loop = (isinstance(parent, ast.For) and parent.iter is not node or
                              isinstance(parent, ast.While) and parent.test is not node)
        if is_containing_loop:
            result.append(parent)

        elif isinstance(parent, (ast.ListComp,
                                 ast.GeneratorExp,
                                 ast.DictComp,
                                 ast.SetComp)):
            if isinstance(parent, ast.DictComp):
                is_comprehension_element = node in (parent.key, parent.value)
            else:
                is_comprehension_element = node is parent.elt
            if is_comprehension_element:
                result.extend(reversed(parent.generators))

            if node in parent.generators:
                result.extend(reversed(list(takewhile(lambda n: n != node, parent.generators))))

        elif isinstance(parent, ast.comprehension) and node in parent.ifs:
            result.append(parent)

        node = parent

    result.reverse()
    return tuple(result)



from __future__ import print_function, division, absolute_import

from future import standard_library

standard_library.install_aliases()
import sys

from flask import Flask, request
from flask.templating import render_template
from flask_humanize import Humanize
from werkzeug.routing import PathConverter

from birdseye.db import Call, Function, Session
from birdseye.utils import all_file_paths, short_path

app = Flask('birdseye')
Humanize(app)


class FileConverter(PathConverter):
    regex = '.*?'


app.url_map.converters['file'] = FileConverter


@app.route('/')
def index():
    files = sorted(all_file_paths())
    files = zip(files, [short_path(f, files) for f in files])
    return render_template('index.html',
                           files=files)


@app.route('/file/<file:path>')
def file_view(path):
    return render_template('file.html',
                           funcs=sorted(Session().query(Function.name).filter_by(file=path).distinct()),
                           full_path=path,
                           short_path=short_path(path))


@app.route('/file/<file:path>/function/<func_name>')
def func_view(path, func_name):
    session = Session()
    query = (session.query(Call, Function)
             .join(Function)
             .filter_by(file=path, name=func_name)
             .order_by(Call.start_time.desc())
             [:200])
    if query:
        func = query[0][1]
        calls = [p[0] for p in query]
    else:
        func = session.query(Function).filter_by(file=path, name=func_name)[0]
        calls = None

    return render_template('function.html',
                           func=func,
                           calls=calls)


@app.route('/call/<call_id>')
def call_view(call_id):
    call = Session().query(Call).filter_by(id=call_id).one()
    func = call.function
    return render_template('call.html',
                           call=call,
                           func=func)


@app.route('/kill', methods=['POST'])
def kill():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()
    return 'Server shutting down...'


def main():
    try:
        port = int(sys.argv[1])
    except IndexError:
        port = 7777

    app.run(debug=True, port=port)


if __name__ == '__main__':
    main()
